const path = require('node:path')
const spawn = require('cross-spawn')
const fs = require('fs-extra')
const chalk = require('chalk')
const globby = require('globby')
const download = require('download')
const Spinnies = require('spinnies')

const {
  PATH_DIST,
  PATH_TMP,
  PATH_REPO_DOC,
  PATH_REPO_DOC_CONFIG,
  PATH_REPO_DOC_PUBLIC,
  DOC_REPO, USE_CNPM
} = require('./config')

// the process to install dependencies
let installProcess
// the process to build site
let buildSiteProcess
// the process to build doc
let buildProcess

// spinners
const spinners = new Spinnies({
  succeedPrefix : '✔️',
  failPrefix: '❌'
})

const SPINNER_MAIN = 'spinner-main'

/**
 * To clone the documentation repository
 */
async function cloneDocRepo() {
  spinners.add(SPINNER_MAIN, { color: 'yellow' })

  spinners.update(SPINNER_MAIN, {
    text: 'Cloning doc repo from ' + DOC_REPO + '...'
  })

  const clonePromise = download(
    DOC_REPO + '/archive/master.tar.gz',
    PATH_REPO_DOC,
    { extract: true }
  )
  try {
    await clonePromise

    const decompressedDir = path.resolve(PATH_REPO_DOC, 'echarts-doc-master')
    const files = await globby('**/*', {
      cwd: decompressedDir,
      // dot: true
    })
    await Promise.all(
      files.map(file =>
        fs.move(
          path.join(decompressedDir, file),
          path.join(PATH_REPO_DOC, file),
          { overwrite: true }
        )
      )
    )
    await fs.remove(decompressedDir)

    spinners.succeed(SPINNER_MAIN, {
      text: 'Clone doc repo, done.'
    })
    console.log()
  }
  catch (e) {
    spinners.fail(SPINNER_MAIN, {
      color: 'red'
    })

    throw e
  }
}

async function install() {
  spinners.add(SPINNER_MAIN, {
    color: 'yellow',
    text: 'Installing dependencies...'
  })

  const installPromise = new Promise((resolve, reject) => {
    installProcess = spawn(
      USE_CNPM ? 'cnpm' : 'npm', ['ci'],
      {
        cwd: PATH_REPO_DOC,
        stdio: 'ignore',
        windowsHide: true,
        detached: false
      }
    )

    installProcess.on('close', (code) => {
      code
        ? reject('failed to install dependencies with code ' + code)
        : resolve()
    })

    installProcess.on('error', reject)
  })

  try {
    await installPromise

    spinners.succeed(SPINNER_MAIN, {
      text: 'Install dependencies, done.'
    })
    console.log()
  }
  catch (e) {
    spinners.fail(SPINNER_MAIN, {
      color: 'red'
    })
    throw e
  }
}

async function build() {
  spinners.add(SPINNER_MAIN, {
    color: 'yellow',
    text: 'Building...'
  })

  await fs.remove(PATH_DIST)

  const configFileName = 'env.local.js'
  // copy config
  await fs.copy(
    path.resolve(__dirname, `./config/${configFileName}`),
    path.resolve(PATH_REPO_DOC_CONFIG, configFileName)
  )

  if (!process.env.GITHUB_ACTIONS) {
    // FIXME echarts-doc has compatibility issues with node >= 16
    if (+process.version.slice(1).split('.')[0] > 16) {
      process.env.NODE_OPTIONS = '--openssl-legacy-provider'
    }
  }

  const buildPromise = new Promise((resolve, reject) => {
    buildSiteProcess = spawn(
      'npm', ['run', 'build:site'],
      {
        cwd: PATH_REPO_DOC,
        stdio: 'ignore',
        windowsHide: true,
        detached: false
      }
    )

    buildSiteProcess.on('close', (code) => {
      if (code === 0) {
        buildProcess = spawn(
          'node', ['build.js', '--env', 'local'],
          {
            cwd: PATH_REPO_DOC,
            stdio: 'ignore',
            windowsHide: true,
            detached: false
          }
        )
        buildProcess.on('close', (code) => {
          code
            ? reject('failed to build with code ' + code)
            : resolve()
        })

        buildProcess.on('error', reject)
      }
      else {
        reject('failed to build with code ' + code)
      }
    })

    buildSiteProcess.on('error', reject)
  })

  try {
    await buildPromise

    const publicDist = path.resolve(PATH_DIST, './public')
    await Promise.all([
      // copy to dist
      fs.copy(PATH_REPO_DOC_PUBLIC, publicDist),
      // copy index redirect
      fs.copy(
        path.resolve(__dirname, 'config/index-redirect.html'),
        path.resolve(PATH_DIST, 'index.html')
      )
    ])

    const options = {
      cwd: publicDist,
      absolute: true,
      onlyFiles: true
    }

    const deleteFiles = await globby(['**/*-content.html'], options)
    deleteFiles.forEach(file => fs.removeSync(file))

    const viewOnlineJS = require('./config/view-online')
    // inject view-online script
    const htmls = await globby(['**/*.html'], options)
    htmls.forEach(html => {
      let content = fs.readFileSync(html, { encoding: 'utf8' })
      content = content.replace('<\/body>', viewOnlineJS(html.indexOf('/zh/') !== -1 ? 'zh' : 'en') + '</body>')
      fs.writeFileSync(html, content, { encoding: 'utf8' })
    })

    spinners.succeed(SPINNER_MAIN, {
      text: 'Build, done.'
    })
    console.log()
  }
  catch (e) {
    spinners.fail(SPINNER_MAIN, {
      color: 'red'
    })
    throw e
  }
}

/**
 * clean up temporary files
 */
function cleanup() {
  try {
    // remove tmp files
    fs.removeSync(PATH_TMP)
  } catch (e) {
    console.error(chalk.red('failed to clean up'))
    throw e
  }
}

process.on('beforeExit', code => {
  cleanup()
  process.exit(code)
})

process.on('uncaughtException', e => {
  console.error(chalk.red('An uncaught error occurred'))
  console.error(chalk.red(e))
  process.exit(-2)
})

module.exports = async function run() {
  try {
    // cleanup first
    cleanup()

    // clone doc repo
    await cloneDocRepo()

    // install necessary dependencies
    await install()

    // download
    const download = require('./build/download')
    const fileMappings = await download()

    // do some modifications
    const modify = require('./build/modify')
    await modify(fileMappings)

    // build
    await build()

  } catch (e) {
    console.error(chalk.red(e))
    process.exit(-1)
  }
}
